/**
  *  \file client/cscreen.cc
  *  \brief Control Screens, and Related Things
  *
  *  This contains the main classes for control screens.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chartusr.h"
#include "client/cscreen.h"
#include "client/dlg-list.h"
#include "client/playerscr.h"
#include "client/screens.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/clip.h"
#include "gfx/palette.h"
#include "resmgr/resmgr.h"
#include "ui/layout.h"
#include "ui/skin.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/nls.h"

//#define BLUESKIN
#ifndef BLUESKIN
#  define ICOLOR (COLOR_SHIELD+3)
#  define IFILL GRAY50_FILL
#else
#  define ICOLOR (COLOR_SHIELD+1)
#  define IFILL SOLID_FILL
#  define FCOLOR (COLOR_SHIELD+5)
#endif

/********************************* WTile *********************************/

/** \class WTile
    \brief Widget displaying certain data (hooray.)

    Tiles can display the following stuff:
    - a constant heading
    - some fixed data
    - some variable data

    This class provides an interface intended to make display of such
    stuff easier. It automatically manages update events. In addition,
    a tile can request to have some associated controls (buttons).

    (In PCC 1.x, the distinguishing property of a "Tile" was that it
    was freely moveable. In PCC 2, all widgets are moveable.) */

/** Standard constructor.
    \param asize size of this tile
    \param id    widget Id of this tile
    \param when  when this tile wants to be updated (flags). Zero means
    update on all events. Otherwise, the tile will only be redrawn when
    one of the given flags is set in the update event. */
WTile::WTile(GfxPoint asize, int id, int when)
    : UIWidget(id), size(asize), update_when(when), state(sContentOnly)
{
    computeCoordinates();
}

/** Standard constructor. This sets the widget Id to zero.
    \param wi,he size of this tile
    \param when see other constructor */
WTile::WTile(int wi, int he, int when)
    : UIWidget(0), size(wi, he), update_when(when), state(sContentOnly)
{
    computeCoordinates();
}

/** Draw the tile. Do not override in WTile descendants. */
void
WTile::draw(GfxCanvas& can, UpdateInfo& ui)
{
    if (state != sContentOnly) {
        getSkin().startDraw();
        setColor(UIColor::tc_White);
        getSkin().setSolid(false);
        getSkin().outTextF(can, FONT_TITLE, extent.x, extent.y, extent.w, heading);
    }
    if (state != sHeadingOnly) {
        getSkin().startDraw();
        drawFixed(can);
        getSkin().startDraw();
        drawData(can);
    }
    ui.update(getExtent());
}

/** Update event handler. This function is hooked into the
    WControlScreen::sig_changed signal. */
void
WTile::update(int n)
{
    if (!n || !update_when || (n & update_when))
        update();
}

/** Update. This function redraws the /variable data/ portion of the
    tile. */
void
WTile::update()
{
    ObjectHolder h;
    GfxCanvas& can = getOwner()->getCanvasFor(this, h);
    UISkin& s = getSkin();

    s.startDraw();
    s.setSolid();
    can.lock();
    drawData(can);
    can.unlock();
    can.update(extent);
}

/** Draw fixed data. Override this in derived classes to display data
    which does not change during the lifetime of this tile. */
void
WTile::drawFixed(GfxCanvas& /*can*/)
{ }

/** Draw variable data. Override this in derived classes to display
    data which does change during the lifetime of the tile. */
void
WTile::drawData(GfxCanvas& /*can*/)
{ }

/** Create controls belonging to this tile. Controls will be added to
    the group. The group will not have a layout manager, so use the
    two-argument add() version. You can use the provided ObjectHolder. 
    For convenience, you can use the WTile::addControl() function
    which automatically adjusts the coordinates. */
void
WTile::addControls(UIGroup& /*g*/, ObjectHolder& /* holder */)
{ }

/** Add control to group. You can use this function to add a control
    to a group in response to an addControls() request.
    \param g the group passed to addControls()
    \param r relative position
    \param w the widget you want to add
    \param rel_pos if true, the coordinates are relative to your
    data area. If false (default), coordinates are relative to
    the whole tile. */
void
WTile::addControl(UIGroup& g, GfxRect r, UIWidget& w, bool rel_pos)
{
    r.moveBy(extent.x, extent.y);
    if (rel_pos)
        r.moveBy(0, y_origin);
    g.add(r, w);
}

/** Draw background. Derived drawData()/drawFixed() can use this to clear
    out regions of the background. \c r is in relative coordinates. */
void
WTile::drawBackground(GfxCanvas& can, GfxRect r)
{
    r.moveBy(extent.x, extent.y + y_origin);
    r.intersect(extent);
    getSkin().drawBackground(can, r);
}

/** Get Layout Information. Do not override in WTile descendants. */
void
WTile::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint size = this->size;
    if (state == sHeadingOnly)
        size.y = 0;
    size.y += y_origin;

    info.flags = 0;
    info.min_size = info.pref_size = size;
}

bool
WTile::handleEvent(const UIEvent& e, bool second_pass)
{
    return UIWidget::handleEvent(e, second_pass);
}

/** Get size of data area, with margins. This returns the dimensions
    of the data area you are expected to fill, in absolute coordinates.
    \param lmargin left margin (number of pixels to remove at the left)
    \param rmargin right margin */
GfxRect
WTile::getInnerSize(int lmargin, int rmargin)
{
    GfxRect rv = getExtent();
    rv.x += lmargin;
    rv.w -= lmargin+rmargin;
    rv.y += y_origin;
    rv.h -= y_origin;
    return rv;
}

/** Set display state.
    \todo Currently, this only works reliably when the tile is not
    yet in a group. It does not resize the tile. */
void
WTile::setDisplayState(State state)
{
    if (state == this->state)
        return;
    this->state = state;
    computeCoordinates();
    drawWidget();
    // FIXME: if widget is *really* already displayed, we probably
    // should tell our parent to resize us
}

/** Set heading of the tile. */
WTile&
WTile::setHeading(string_t heading)
{
    this->heading = heading;
    drawWidget();
    return *this;
}

/** Compute coordinates, after a change to layout parameters. \internal */
void
WTile::computeCoordinates()
{
    if (state == sContentOnly) {
        y_origin = 0;
    } else {
        y_origin = font_heights[FONT_TITLE];
    }
}

/************************** WControlScreenSkin ***************************/

/** \class WControlScreenSkin
    \brief Skin for control screens */

const UIColorSet&
WControlScreenSkin::getColors() const
{
    return DARK_COLOR_SET;
}

#ifdef BLUESKIN
void
WControlScreenSkin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    if (isSolid()) {
        int width = fonts[f]->getTextWidth(text);
        drawBackground(can, GfxRect(x - width*parm.just_x/2, y, width, font_heights[f]));
    }
    parm.bgcolor = GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outText(can, parm, x, y, text);
}

void
WControlScreenSkin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    if (isSolid())
        drawBackground(can, GfxRect(x - lim*parm.just_x/2, y, lim, font_heights[f]));
    parm.bgcolor = GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outTextF(can, parm, x, y, lim, text);
}

void
WControlScreenSkin::drawBackground(GfxCanvas& can, GfxRect r)
{
    can.drawBar(r, standard_colors[ICOLOR], standard_colors[0], IFILL, GFX_ALPHA_OPAQUE);
}
#else
void
WControlScreenSkin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    parm.bgcolor = isSolid() ? standard_colors[COLOR_DARK] : GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outText(can, parm, x, y, text);
}

void
WControlScreenSkin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    parm.bgcolor = isSolid() ? standard_colors[COLOR_DARK] : GFX_TRANSPARENT;
    parm.font = fonts[f];
    ::outTextF(can, parm, x, y, lim, text);
}

void
WControlScreenSkin::drawBackground(GfxCanvas& can, GfxRect r)
{
    drawSolidBar(can, r, standard_colors[COLOR_DARK]);
}
#endif

/***************************** WControlScreen ****************************/

/*
  Layout:

    ,-----------------------------------.
    |              ,------------------. |   <-- borders are 5 pixels
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              `------------------' |  <-- tile_new_y  \ 3 lines for
    |              [                  ] |  ___ chart_y2    / scanner widget
    |                                   |
    |                                   |    8 lines free here for tiles
    |                                   |
    |                                   |
    `-----------------------------------'  ---

    |  320 pixels  |
                chart_x1
 */

/** Constructor.
    \param atype object type we're displaying. Used for iteration.
    \param screen_id this screen's identifier.
    \param update_when update flags. When one of the bits set here is in our
    object's dirty flags, we need to call update().

    Checklist for derived constructors:
    - create all tiles and add them with addTile(). You should include a
      WControlScreenHeaderTile.
    - create starchart widget and hand it to addChart()
    - connect updateId() to the correct signal */
WControlScreen::WControlScreen(GObjectType& atype, int screen_id, int update_when)
    : scanner(),
      screen_id(screen_id),
      update_when(update_when),
      pos_changed_cookie(atype.sig_pos_changed.add(this, &WControlScreen::onPosChanged)),
      type(atype), pic_id(0), pos(2000, 2000),
      chart(0),
      controls()
{
    setSkin(skin);
    setExtent(GfxRect(ui_screen->getSystemSurface()), true);
    setGeometry();

    controls.setExtent(extent);
    add(controls);

    const int scanner_height = 2*font_heights[FONT_NORMAL];
    add(GfxRect(chart_x1+2,                chart_y2 + 5,
                extent.w - chart_x1 - 5-2, scanner_height),
        scanner);
}

WControlScreen::~WControlScreen()
{
    type.sig_pos_changed.remove(pos_changed_cookie);
}

/** Compute geometry parameters. */
void
WControlScreen::setGeometry()
{
    tile_width = 315;           // FIXME: should probably be related to font size somehow
    tile_new_y = extent.h - 8*font_heights[FONT_NORMAL];
    tile_y     = 5;
    tile_x     = 5;
    chart_y2   = tile_new_y - 3*font_heights[FONT_NORMAL];
    chart_x1   = tile_width + tile_x + 5;
}

/** Set data to display in the title area. Call this from your
    update() implementation. */
void
WControlScreen::setData(ResID apic, GPoint apos,
                        const string_t& atit, const string_t& adesc)
{
    pic_id = apic;
    pos    = apos;
    title  = atit;
    desc   = adesc;
    chart->setCenterPosition(apos);
}

/** Add a tile, and give it a title. */
void
WControlScreen::addTile(WTile& tile, const char* title)
{
    tile.setHeading(title);
    tile.setDisplayState(WTile::sNormal);
    addTile(tile);
}

/** Add a tile. */
void
WControlScreen::addTile(WTile& tile)
{
    UIWidget::LayoutInfo info;
    tile.getLayoutInfo(info);

    if (tile_y + info.pref_size.y > extent.h - 5) {
        /* tile does not fit in this column. Would it fit when we
           remove the heading? */
        if (tile.getDisplayState() == WTile::sNormal) {
            tile.setDisplayState(WTile::sContentOnly);
            tile.getLayoutInfo(info);
            if (tile_y + info.pref_size.y <= extent.h - 5)
                goto success;
            tile.setDisplayState(WTile::sNormal);
            tile.getLayoutInfo(info);
        }

        /* start new column */
        tile_y = tile_new_y;
        tile_x += tile_width;

        /* does it fit there? */
        if (tile_y + info.pref_size.y > extent.h - 5) {
            /* nope. Remove heading if there, but we cannot do
               anything else so let it be. */
            if (tile.getDisplayState() == WTile::sNormal) {
                tile.setDisplayState(WTile::sContentOnly);
                tile.getLayoutInfo(info);
            }
        }
    }
 success:
    tile.setExtent(GfxRect(tile_x, tile_y, tile_width, info.pref_size.y), true);
    tile.addControls(controls, holder);
    tile_y += info.pref_size.y + 2;

    sig_changed.add(&tile, (void (WTile::*)(int)) &WTile::update);
    add(tile);
}

/** Add starchart widget. */
void
WControlScreen::addChart(WScannerChartWidget& chart)
{
    this->chart = &chart;
    add(GfxRect(chart_x1, 5, extent.w - chart_x1 - 5, chart_y2 - 5), chart);
    scanner.setChart(chart);
}

/** Go to object #id. This one is tolerant in that it silently ignores
    invalid Ids. */
void
WControlScreen::setCurrent(int id)
{
    if (id == type.getCurrent() || !type.isExistant(id))
        return;
    type.setCurrent(id);
}

void
WControlScreen::drawInterior(GfxCanvas& can, UpdateInfo& ui)
{
    // FIXME: optimize!
    //drawSolidBar(can, extent, standard_colors[COLOR_DARK]);
    getSkin().startDraw();
    getSkin().drawBackground(can, extent);

#ifdef BLUESKIN
    getSkin().getParameters().color = standard_colors[FCOLOR];
    drawRectangle(can, getSkin(), GfxRect(extent.x, extent.y+1, extent.w, extent.h-2));
    drawRectangle(can, getSkin(), GfxRect(extent.x+1, extent.y, extent.w-2, extent.h));

    drawRectangle(can, getSkin(), GfxRect(chart_x1-1, 5-2,        extent.w - chart_x1 - 3, chart_y2 - 5+4));
    drawRectangle(can, getSkin(), GfxRect(chart_x1-2, 5-1,        extent.w - chart_x1 - 1, chart_y2 - 5+2));

    drawRectangle(can, getSkin(), GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
#else
    drawFrameUp(can, extent);
    drawFrameDown(can, GfxRect(chart_x1-1, 5-1,        extent.w - chart_x1 - 3, chart_y2 - 5+2));
    drawFrameDown(can, GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
#endif
    ui.update(extent);
}

bool
WControlScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch(e.key.code) {
         case SDLK_PAGEUP:
         case '-':
         case vk_WheelUp:
            setCurrent(type.getPrev(type.getCurrent(), false));
            return true;
         case SDLK_PAGEDOWN:
         case '+':
         case vk_WheelDown:
            setCurrent(type.getNext(type.getCurrent(), false));
            return true;
         case SDLK_PAGEUP | ss_Ctrl:
         case '-' | ss_Ctrl:
            setCurrent(type.getPrev(type.getCurrent(), true));
            return true;
         case SDLK_PAGEDOWN | ss_Ctrl:
         case '+' | ss_Ctrl:
            setCurrent(type.getNext(type.getCurrent(), true));
            return true;
         case SDLK_ESCAPE:
            stop(0);
            return true;
         case '.':
            type.toggleMark(type.getCurrent());
            type.getTurn().doUpdates();
            return true;
         case SDLK_F1:
            doScreen(1);
            return true;
         case SDLK_F1 + ss_Shift:
         case SDLK_F1 + ss_Ctrl:
            if (chart->isValid()) {
                GObjectId id = doListShips(((e.key.code & ss_Ctrl)
                                            ? chart->getScannerPosition()
                                            : pos),
                                           ls_EarlyExit,
                                           (screen_id==1 || screen_id==11) ? type.getCurrent() : 0,
                                           _("Visual Scanner"),
                                           _("Go to"),
                                           0 /* FIXME */);
                if (id.isValid())
                    putCommand(cm_CS_GotoCS, 1, id.getId());
            }
            return true;
         case SDLK_F2:
            doScreen(2);
            return true;
         case SDLK_F2 + ss_Shift:
            putCommand(cm_CS_GotoCS, 2, type.getTurn().getPlanetAt(chart->getCenterPosition(), true));
            return true;
         case SDLK_F2 + ss_Ctrl:
            putCommand(cm_CS_GotoCS, 2, scanner.getScannedPlanetId(false));
            return true;
         case SDLK_F3:
            doScreen(3);
            return true;
         case SDLK_F3 + ss_Shift:
            putCommand(cm_CS_GotoCS, 3, type.getTurn().getPlanetAt(chart->getCenterPosition(), true));
            return true;
         case SDLK_F3 + ss_Ctrl:
         case SDLK_F8 + ss_Ctrl:
            putCommand(cm_CS_GotoCS, 3, scanner.getScannedPlanetId(false));
            return true;
         case SDLK_F4:
         case SDLK_F4 + ss_Shift:
            if (chart->isValid())
                setStarchartPosition(chart->getCenterPosition());
            stop(4);
            return true;
         case SDLK_F4 + ss_Ctrl:
            if (chart->isValid()) {
                setStarchartPosition(chart->getScannerPosition());
                stop(4);
            }
            return true;
         case SDLK_F5:
         case SDLK_F5 + ss_Shift:
            if (int pid = getDisplayedTurn().getPlanetAt(chart->getCenterPosition()))
                doPlanetScan(pid);
            return true;
         case SDLK_F5 + ss_Ctrl:
            if (int pid = getDisplayedTurn().getPlanetAt(chart->getScannerPosition()))
                doPlanetScan(pid);
            return true;
         case 'l':
            if (chart->isValid()) {
                GObjectId id = doListShips(chart->getScannerPosition(),
                                           ls_IncludeForeignShips,
                                           0,
                                           _("Visual Scanner"),
                                           _("Go to"),
                                           0 /* FIXME */);
                if (id.isValid())
                    putCommand(cm_CS_GotoCS, 1, id.getId());
            }
            return true;
        }
    } else if (e.type == ev_Command) {
        switch(e.cmd.code) {
         case cm_CS_GotoCS:
            if (e.cmd.info == screen_id) {
                if (type.isExistant(e.cmd.info2))
                    setCurrent(e.cmd.info2);
            } else if (e.cmd.info == 0) {
                stop(0);
            } else {
                GObjectType& t = ::getObjectType(e.cmd.info);
                if (t.isExistant(e.cmd.info2)) {
                    t.setCurrent(e.cmd.info2);
                    stop(e.cmd.info);
                }
            }
            return true;
        }
    }
    return UIGroup::handleEvent(e, second_pass);
}

int
WControlScreen::run()
{
    update();
    ui_root->add(*this);
    int rv = UIGroup::run();
    ui_root->close(*this);
    return rv;
}

/** Selection-change handler. This is called when the user picks a
    different object on this object's type. */
void
WControlScreen::onPosChanged(int /*id*/)
{
    update();
    sig_changed.raise(0);
}

/** Object-changed handler. This is called when doUpdates() was called
    and an object of our type was modified.

    Derived constructors must connect this to the correct signal. */
void
WControlScreen::updateId(int id)
{
    if (id == type.getCurrent()) {
        int df = type.getDirtyFlags(id);
        if (df & update_when)
            update();
        sig_changed.raise(df);
    }
}

/************************ WControlScreenHeaderTile ***********************/

/*! \class WControlScreenHeaderTile
    \brief Header of a Control Screen

    Displays the top-most part of a control screen: name, description,
    picture, buttons. */

/** Constructor.
    \param acon control screen
    \param buttons the buttons you want to be displayed
    \param update flags when this tile should be redrawn */
WControlScreenHeaderTile::WControlScreenHeaderTile(WControlScreen& acon, int buttons, int update)
      // FIXME: magic numbers
    : WTile(GfxPoint(315, 145), update), con(acon), buttons(buttons)
{ }

void
WControlScreenHeaderTile::drawData(GfxCanvas& can)
{
    // FIXME: magic numbers
    int wi = (buttons & bName) ? extent.w-20 : extent.w;
    setColor(UIColor::tc_White);
    outTextF(can, FONT_TITLE, 0, 0, wi, con.title);
    setColor(UIColor::tc_Yellow);
    outTextF(can, FONT_NORMAL, 0, 25, extent.w, con.desc);

    drawFrameDown(can, GfxRect(extent.x + 108, extent.y + 45, 107, 95));

    Ptr<GfxPixmap> pix = getResourcePixmap(con.pic_id, rc_Normal);
    if (pix)
        pix->blitSized(can, extent.x + 109, extent.y + 46, 105, 93);

    drawColorFrame(can, GfxRect(extent.x + 108, extent.y + 45, 107, 95),
                   con.getType().isMarked(con.getType().getCurrent()) ? cf_Yellow : cf_None,
                   false);
}

void
WControlScreenHeaderTile::addControls(UIGroup& g, ObjectHolder& holder)
{
    addControl(g, GfxRect( 5,  45, 30, 25),
               holder.add(new UIButton("F1", SDLK_F1)));
    addControl(g, GfxRect(40,  45, 30, 25),
               holder.add(new UIButton("F2", SDLK_F2)));
    addControl(g, GfxRect( 5,  75, 30, 25),
               holder.add(new UIButton("F3", SDLK_F3)));
    addControl(g, GfxRect(40,  75, 30, 25),
               holder.add(new UIButton("F4", SDLK_F4)));
    addControl(g, GfxRect( 5, 105, 30, 25),
               holder.add(new UIButton("F6", SDLK_F6)));
    addControl(g, GfxRect(40, 105, 30, 25),
               holder.add(new UIButton("F7", SDLK_F7)));
    addControl(g, GfxRect(265, 114, 45, 25),
               holder.add(new UIButton("ESC", SDLK_ESCAPE)));
    addControl(g, GfxRect(230, 114, 25, 25),
               holder.add(new UIButton("H", 'h')));
    addControl(g, GfxRect(285, 45, 25, 25),
               holder.add(new UIButton(cm_CS_Xchg, "X", 'x', UIButton::aKey)));
    addControl(g, GfxRect(75, 45, 20, 25),
               holder.add(new UIButton("\036", '-')));
    addControl(g, GfxRect(75, 75, 20, 25),
               holder.add(new UIButton("\037", '+')));
    if (buttons & bSend)
        addControl(g, GfxRect(75, 105, 20, 25),
                   holder.add(new UIButton("I", 'i')));
    if (buttons & bAuto)
        addControl(g, GfxRect(230, 45, 50, 25),
                   holder.add(new UIButton("Auto", SDLK_RETURN)));
    if (buttons & bCycle)
        addControl(g, GfxRect(275, 75, 35, 25),
                   holder.add(new UIButton("\035", SDLK_TAB)));
    if (buttons & bAdd)
        addControl(g, GfxRect(230, 75, 40, 25),
                   holder.add(new UIButton("Add", SDLK_INSERT)));
    if (buttons & bName)
        addControl(g, GfxRect(extent.w-20, 0, 20, 20),
                   holder.add(new UIButton("N", 'n')));
}

/****************************** WHeadingTile *****************************/

/*! \class WHeadingTile
    \brief Section divider

    Displays a constant text string in big font, as a divider between
    several sections on a control screen.

    This tile is now superfluous, as every tile can display a heading. */

WHeadingTile::WHeadingTile(const string_t& s)
    : WTile(GfxPoint(315, font_heights[FONT_TITLE])), text(s)
{ }

WHeadingTile::WHeadingTile(const char* s)
    : WTile(GfxPoint(315, font_heights[FONT_TITLE])), text(s)
{ }

void
WHeadingTile::drawFixed(GfxCanvas& can)
{
    setColor(UIColor::tc_White);
    outTextF(can, FONT_TITLE, 0, 0, extent.w, text);
}

/******************************** WScanner *******************************/

/** \class WScanner
    \brief Scanner Evaluation

    This widget displays the result of a scan (distance, name of
    object pointed-to). */

WScanner::WScanner()
    : chart(0)
{ }

WScanner::~WScanner()
{ }

void
WScanner::setChart(WScannerChartWidget& achart)
{
    ASSERT(!chart);
    chart = &achart;
    chart->sig_moved.add(this, &WScanner::update);
    update();
}

void
WScanner::update()
{
    if (!chart || !chart->isValid()) {
        sid = xpid = gpid = 0;
        if (owner)
            owner->setStateOf(cm_CS_Xchg, st_Disabled, true);
        return;
    }

    sid  = getDisplayedTurn().getShipAt(chart->getScannerPosition());
    xpid = getDisplayedTurn().getPlanetAt(chart->getScannerPosition());
    if (xpid)
        gpid = xpid;
    else
        gpid = getDisplayedTurn().getGravityPlanetAt(chart->getScannerPosition());
    if (owner)
        owner->setStateOf(cm_CS_Xchg, st_Disabled, (xpid == 0 && gpid == 0 && sid == 0));

    drawWidget(false);
}

void
WScanner::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UITableWriter w(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);
    string_t scan, dist;

    w.newLColumn();
    w.writeFixedCell(UIColor::tc_Static, _("Scan:"));
    w.writeFixedCell(UIColor::tc_Static, _("Distance:"));

    w.newLColumn();
    if (chart) {
        if (!chart->isValid()) {
            scan = dist = "...";
        } else {
            if (xpid > 0)
                scan = getDisplayedTurn().getPlanet(xpid).getName();
            else if (sid > 0)
                scan = getDisplayedTurn().getShip(sid).getFullName();
            else
                scan = format(_("Deep Space (%d,%d)"),
                              chart->getScannerPosition().x,
                              chart->getScannerPosition().y);
            dist = format(_("%.2f ly"),
                          std::sqrt(double(distanceSquaredRaw(chart->getScannerPosition(),
                                                              chart->getCenterPosition()))));
        }
    }
    w.writeVariableCell(UIColor::tc_Green, scan);
    w.writeVariableCell(UIColor::tc_Green, dist);
    w.finish();
    ui.update(getExtent());
}

void
WScanner::getLayoutInfo(LayoutInfo& info)
{
    // FIXME?
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(2*font_heights[FONT_NORMAL], 310);
}
