/**
  *  \file client/cscreen.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_CSCREEN_H
#define PCC_V2_CLIENT_CSCREEN_H

#include "client/chart.h"
#include "cpluslib/smartptr.h"
#include "game/objtypes.h"
#include "resmgr/resmgr.h"
#include "ui/commands.h"
#include "ui/ui.h"
#include "cpluslib/holder.h"
#include "ui/skin.h"

enum {
    cm_CS_ToggleSel = cm_ControlScreenFirst,
    cm_CS_Xchg,                    // "X"
    cm_CS_ReturnFromCS,
    cm_CS_ScannerDouble,
    cm_CS_MoveManually,            // "Y"
    cm_CS_ListShips,               // "L"
    cm_CS_GotoCS                   // args are screen/object id
//     cm_CS_Reset,                // "Alt-R"
//     cm_CS_Console,              // "Alt-C"
//     cm_CS_Taskmanager,          // "Alt-Q"
    // cm_EditorUpdated
//     cm_ToggleAuto,              // "Enter"

//     cm_CS_FirstUser = 2000
};

class WControlScreen;

class WTile : public UIWidget {
    GfxPoint size;
    int      update_when;
    string_t heading;
    int      y_origin;
 public:
    enum State {
        sNormal,
        sHeadingOnly,
        sContentOnly
    };
 private:
    State state;
    void computeCoordinates();
 protected:
    virtual void drawFixed(GfxCanvas& can);
    virtual void drawData(GfxCanvas& can);
    /** Set text color. Shorthand for setting the color in the skin. */
    void setColor(UISkin::Color c)
        { getSkin().setColor(c); }
    /** Output text. Shorthand for UISkin::outText. In addition, this
        automatically does coordinate transformation. */
    void outText(GfxCanvas& can, int f, int x, int y, string_t text)
        { getSkin().outText(can, f, x + extent.x, y + extent.y + y_origin, text); }
    /** Output text. Shorthand for UISkin::outTextF. In addition, this
        automatically does coordinate transformation. */
    void outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
        { getSkin().outTextF(can, f, x + extent.x, y + extent.y + y_origin, lim, text); }
    void drawBackground(GfxCanvas& can, GfxRect r);
    void addControl(UIGroup&, GfxRect r, UIWidget& w, bool rel_pos = false);
 public:
    WTile(GfxPoint size, int aid = 0, int awhen = 0);
    WTile(int wi, int he, int when = 0);
    void draw(GfxCanvas& can, UpdateInfo& ui);
    void update();
    void update(int);
    void getLayoutInfo(LayoutInfo& info);
    bool handleEvent(const UIEvent& e, bool second_pass);
    void setDisplayState(State st);
    State getDisplayState() const
        { return state; }
    WTile& setHeading(string_t heading);
    const string_t getHeading() const
        { return heading; }

    virtual void addControls(UIGroup&, ObjectHolder&);
    GfxRect getInnerSize(int lmargin, int rmargin);
};

class WObjectSelectionTile : public WTile {
    GObjectType& type;
    int current;
 public:
    WObjectSelectionTile(GObjectType& atype);
    void         select(int id);
    bool         handleEvent(const UIEvent& e, bool second_pass);

    GObjectType& getType() const { return type; }
    int          getCurrent() const { return current; }
};

class WChartWidget : public UIWidget {
    GChartViewport viewport;
 public:
    WChartWidget();
    ~WChartWidget();
    void draw(GfxCanvas& can, UpdateInfo& ui);
    virtual void setCenterPosition(GPoint pt);

    virtual void drawPre(GfxCanvas& can);
    virtual void drawPost(GfxCanvas& can);

    GPoint getCenterPosition() const { return viewport.center; }
    bool   isValid() const { return true; } // FIXME:  we need some way to disable a chart widget

    const GChartViewport& getViewport() const { return viewport; }
    bool  zoomIn();
    bool  zoomOut();
};

class WScannerChartWidget : public WChartWidget {
    GPoint      scan;
    // FIXME: bubble stuff
    void drawScannerXor(GfxCanvas& can);
 protected:
    virtual void drawScanner(GfxCanvas& can);
    virtual void sync(bool dblclick);
    static void drawEndBig(GfxCanvas& can, GfxParameters& parm, int x, int y);
    static void drawEndSmall(GfxCanvas& can, GfxParameters& parm, int x, int y);

    void moveKey(int dx, int dy);
    bool handleKey(UIEvent::Key k);
 public:
    WScannerChartWidget();
    ~WScannerChartWidget();
    void drawPre(GfxCanvas& can);
    void drawPost(GfxCanvas& can);
    void setScannerPosition(GPoint pt, bool limit);
    void setCenterPosition(GPoint pt);
    bool handleEvent(const UIEvent& e, bool second_pass);

    GPoint getScannerPosition() const { return scan; }

    void getLayoutInfo(LayoutInfo& info);

    enum {
        sc_StyleChange  = st_Last*2,  ///< user can changes style (C-M-letter)
        sc_Zoomable     = st_Last*4,  ///< user can zoom (+/-)
        sc_Keyboard     = st_Last*8,  ///< user can use keyboard
        sc_ShipsOnly    = st_Last*16, ///< only lock on ships
        sc_KeyboardMode = st_Last*32  ///< keyboard mode (Y) active
    };

    Signal1<int> sig_moved;
};

class WScanner : public UIWidget {
    WScannerChartWidget* chart;
    int sid, xpid, gpid;
 public:
    WScanner();
    ~WScanner();

    void setChart(WScannerChartWidget& achart);
    void update();
    void draw(GfxCanvas& can, UpdateInfo& ui);
    void getLayoutInfo(LayoutInfo& info);

    int getScannedPlanetId(bool exact) const { return exact ? xpid : gpid; }
    int getScannedShipId() const { return sid; }
    WScannerChartWidget* getChart() const { return chart; }
};

class WControlScreenSkin : public UISkin {
 public:
    const UIColorSet& getColors() const;
    void  outText(GfxCanvas& can,  int f, int x, int y, string_t text);
    void  outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text);
    void  drawBackground(GfxCanvas& can, GfxRect r);
};

class WControlScreen : public UIGroup {
    WControlScreenSkin skin;
    WScanner      scanner;
    int           screen_id;    ///< Screen Id (function key number).
    int           update_when;  ///< Update header when these bits in object change.

    void*         pos_changed_cookie;

    void setGeometry();
    void onPosChanged(int);
 protected:
    GObjectType&  type;
 public:
    ResID         pic_id;
    GPoint        pos;
    string_t      title, desc;

    // FIXME: I don't actually like that this parameter is 'int'.
    // Usually, an 'int' is an object identifier.
    Signal1<int>  sig_changed;  ///< Raised when displayed object changes. Parameter is bitmask.

 private:
    /* tile positioning */
    int   tile_width;       // width of one column
    int   tile_y;           // current y
    int   tile_x;
    int   tile_new_y;       // y for new columns
    int   chart_y2;
    int   chart_x1;

    WScannerChartWidget*      chart;
    UIGroup                   controls;
 protected:
    ObjectHolder holder;

    void addChart(WScannerChartWidget& chart);
    void addTile(WTile& tile);
    void addTile(WTile& tile, const char* title);
    void setData(ResID apic, GPoint apos, const string_t& atit, const string_t& adesc);
 public:
    WControlScreen(GObjectType& atype, int screen_id, int update_when);
    ~WControlScreen();

    bool          handleEvent(const UIEvent& e, bool second_pass);
    void          drawInterior(GfxCanvas& can, UpdateInfo& ui);
    void          setCurrent(int id);
    GObjectType&  getType() { return type; }
    GPoint        getScannerPosition() const { return chart->getScannerPosition(); }
    WScanner&     getScanner() { return scanner; }
    int           getScreenId() const { return screen_id; }

    virtual void  update() = 0;

    void          updateId(int id);

    int           run();
};

class WControlScreenHeaderTile : public WTile {
    WControlScreen& con;
    int buttons;
 public:
    WControlScreenHeaderTile(WControlScreen& acon, int buttons, int update);
    void drawData(GfxCanvas& can);
    void addControls(UIGroup& g, ObjectHolder&);

    enum {
        bName  = 1,              // N
        bSend  = 2,              // I
        bAuto  = 4,              // Auto
        bCycle = 8,              // ,->
        bAdd   = 16              // Add (Ins)
    };
};

class WHeadingTile : public WTile {
    string_t text;
 public:
    WHeadingTile(const string_t& s);
    WHeadingTile(const char* s);
    void drawFixed(GfxCanvas& can);
};

int doSelectWindow(WObjectSelectionTile& inter, string_t title);

#endif
