/**
  *  \file client/dlg-ammo.cc
  */

#include "client/dlg-ammo.h"
#include "client/wcost.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "cpluslib/string.h"
#include "game/action.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/widgets.h"
#include "util/nls.h"

namespace {
    class WAmmoList : public UIListbox {
        GStarbaseBuildAmmoAction& p;
     public:
        WAmmoList(GStarbaseBuildAmmoAction& p);
        ~WAmmoList();

        void drawPart(GfxCanvas& can, int from, int to);
        bool isAccessible(int what);
    };

    class WAmmoInfo : public UIWidget {
        GCargoContainer&   transfer;
        GCargoType         observe;
        void*              cookie_with_chocolate_and_nuts;
     public:
        WAmmoInfo(GCargoContainer& transfer, GCargoType observe);
        ~WAmmoInfo();

        void draw(GfxCanvas& can, UpdateInfo& ui);
        void setCargoType(GCargoType observe);
        void getLayoutInfo(LayoutInfo& li);
    };

    class WAmmoWindow : public UIWindow {
        GPlanetTransfer           transfer;
        GStarbaseBuildAmmoAction  action;
        WAmmoList                 list;
        WAmmoInfo                 info;
        WItemCostDisplay          dpy;
        ObjectHolder              h;

        void build(int32 amount);
     public:
        WAmmoWindow(GPlanet& p);

        bool handleEvent(const UIEvent& e, bool second_pass);
        void scrolled(int);
        void enter(int);
        void cancel(int);
    };
};

WAmmoList::WAmmoList(GStarbaseBuildAmmoAction& p)
    : UIListbox(0, NUM_TORPS+1, 18*font_ems[FONT_NORMAL], NUM_TORPS+1),
      p(p)
{ }

WAmmoList::~WAmmoList()
{ }

bool
WAmmoList::isAccessible(int what)
{
    return p.canHave(what+1);
}

void
WAmmoList::drawPart(GfxCanvas& can, int from, int to)
{
    getSkin().startDraw();
    for (int i = from; i < to; ++i) {
        const int n = getTop() + i + 1;
        UIColor::Color c;
        if (int t = p.getTechUpgradeNeeded(n)) {
            if (t > 6 && !allowRegistered(p.getPlanet().getOwner()))
                c = UIColor::tc_Faded;  // FIXME: should be DARK/GRAY, not DARK/DARK
            else
                c = UIColor::tc_Blue;
        } else {
            c = UIColor::tc_Static;
        }
        if (!p.canHave(n))
            c = UIColor::tc_Faded;
        
        string_t what;
        if (p.getAmount(n))
            what += '\007';
        else
            what += ' ';
        if (n <= NUM_TORPS)
            what += getTorp(n).getName();
        else
            what += _("Fighter");

        drawEntry(can, c, i, what);
    }
}

/******************************* WAmmoInfo *******************************/

WAmmoInfo::WAmmoInfo(GCargoContainer& transfer, GCargoType observe)
    : transfer(transfer), observe(observe)
{
    cookie_with_chocolate_and_nuts =
        transfer.sig_changed.add(static_cast<UIWidget*>(this),
                                 (void (UIWidget::*)()) &UIWidget::drawWidget);
}

WAmmoInfo::~WAmmoInfo()
{
    transfer.sig_changed.remove(cookie_with_chocolate_and_nuts);
}

void
WAmmoInfo::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UISkin& s  = getSkin();
    int y      = getExtent().y;
    int32 have = transfer.getCurrent(observe);
    int32 max  = transfer.getMax(observe);

    /* Line 1: "Blafasel Starbase" */
    s.startDraw();
    s.setSolid(false);
    s.setColor(UIColor::tc_Static);
    s.outTextF(can, FONT_TITLE, getExtent().x, y, getExtent().w,
               transfer.getName());
    y += font_heights[FONT_TITLE] + 4;

    /* Line 2: "You have %d <zomething>" */
    s.setSolid();
    s.outTextF(can, FONT_NORMAL, getExtent().x, y, getExtent().w,
               format(observe == el_Fighters
                      ? _("You have %d fighter%!1{s%}.")
                      : _("You have %d torpedo%!1{es%}."),
                      have));
    y += font_heights[FONT_NORMAL];

    /* Line 3: "You have %d left" */
    string_t expl;
    if (have >= max)
        expl = _("This is the maximum amount.");
    else if (max - have <= 5000)
        /* the average player has much fewer torps on a starbase. let
           them rest in peace not knowing there's a limit, most of the
           time. */
        expl = format(_("You have room for %d more."), max - have);
    s.outTextF(can, FONT_NORMAL, getExtent().x, y, getExtent().w, expl);

    ui.update(getExtent());
}

void
WAmmoInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.pref_size = li.min_size = GfxPoint(20 * font_ems[FONT_NORMAL],
                                          2 * font_heights[FONT_NORMAL]
                                          + font_heights[FONT_TITLE] + 5);
}

void
WAmmoInfo::setCargoType(GCargoType observe)
{
    this->observe = observe;
    drawWidget();
}

/****************************** WAmmoWindow ******************************/

WAmmoWindow::WAmmoWindow(GPlanet& p)
    : UIWindow(_("Build Torpedoes/Fighters"), BLUE_WINDOW, COLOR_GRAY),
      transfer(p), action(transfer, transfer), list(action),
      info(transfer, action.getCargoTypeFromIndex(list.getCurrent()+1)),
      dpy(action, true)
{
    setLayout(UIVBoxLayout::instance5);
    add(list);
    add(info);
    add(dpy);

    UIGroup& hg = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIButton& ok = h.add(new UIButton(1000,  _("OK"), SDLK_RETURN, UIButton::aNothing));
    UIButton& esc = h.add(new UIButton(1001, _("Cancel"), SDLK_ESCAPE, UIButton::aNothing));
    hg.add(ok);
    hg.add(esc);
    hg.add(h.add(new UISpacer()));
    add(hg);

    ok.sig_fired.add(this, &WAmmoWindow::enter);
    esc.sig_fired.add(this, &WAmmoWindow::cancel);
    list.sig_moved.add(this, &WAmmoWindow::scrolled);

    pack();
}

void
WAmmoWindow::build(int32 amount)
{
    int index = list.getCurrent()+1;

    // first, if we're trying to build, and don't have sufficient
    // tech, we have to upgrade.
    if (amount == 0)
        return;
    if (amount > 0) {
        if (int tech = action.getTechUpgradeNeeded(index)) {
            // can we upgrade tech?
            if (action.getPlanet().getCargoHandler().isFacilityInUse(GPlanet::pf_TechLevels)) {
                // We can't upgrade tech because someone else is
                // messing with tech levels.
                messageBox(_("Can't upgrade tech level."), _("Tech Upgrade"));
                return;
            }

            GPlanetTransfer tech_pay(action.getPlanet());
            GStarbaseTechUpgradeAction tech_action(action.getPlanet(), tech_pay);
            if (!tech_action.setTech(tl_Torpedo, tech)) {
                messageBox(_("Building this torpedo requires you to upgrade your tech level over Tech 6. "
                             "As an unregistered player, you can't do that."),
                           _("Tech Upgrade"));
                tech_action.cancel();
                tech_pay.cancel();
                return;
            }
            if (!tech_action.isValid()) {
                messageBox(format(_("Building this torpedo requires a tech upgrade. "
                                    "You do not have the %d megacredits required for the upgrade."),
                                  tech_action.getCost()),
                           _("Tech Upgrade"));
                tech_action.cancel();
                tech_pay.cancel();
                return;
            }
            if (UIMessageBox(format(_("Building this torpedo requires tech %d. "
                                      "Do you want to upgrade for %d megacredits?"),
                                    tech,
                                    tech_action.getCost()),
                             _("Tech Upgrade")).doYesNoDialog())
            {
                tech_action.commit();
                tech_pay.commit();
            } else {
                tech_action.cancel();
                tech_pay.cancel();
            }

            // do not immediately buy stuff after upgrading tech
            // FIXME: widgets should figure out itself that it needs to
            // be redrawn
            dpy.drawWidget(false);
            list.drawWidget(false);
            getDisplayedTurn().doUpdates();
            return;
        }
    }

    // okay, we're building stuff
    action.setAmount(index, action.getAmount(index) + amount);
    dpy.drawWidget(false);
    list.drawWidget(false);
}

bool
WAmmoWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case '+':
            build(e.key.arg ? e.key.arg : 1);
            return true;
         case '-':
            build(e.key.arg ? -e.key.arg : -1);
            return true;
         case ss_Ctrl + '+':
            build(100);
            return true;
         case ss_Ctrl + '-':
            build(-100);
            return true;
         case ss_Alt + '+':
            build(10000);
            return true;
         case ss_Alt + '-':
            build(-10000);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

void
WAmmoWindow::scrolled(int)
{
    int what = list.getCurrent() + 1;
    // FIXME: this replicates information from act-ammo.cc. We should
    // have some standardized interface to pass such information around

    WItemCostDisplay::ItemCost c;
    int   tech;

    if (what > 0 && what <= NUM_TORPS) {
        c.t = c.d = c.m = 1;
        c.sup = 0;
        c.mc = getTorp(what).getTorpCost();
        tech = getTorp(what).getTech();
    } else {
        c.t = 3; c.d = 0; c.m = 2; c.sup = 0; c.mc = 100;
        tech = 1;
    }

    dpy.setCost(c);
    dpy.setTech(transfer.getPlanet().getBaseTech(tl_Torpedo), tech);

    info.setCargoType(action.getCargoTypeFromIndex(what));
}

void
WAmmoWindow::enter(int)
{
    if (action.isValid()) {
        action.commit();
        transfer.commit();
        stop(1);
    }
}

void
WAmmoWindow::cancel(int)
{
    action.cancel();
    transfer.cancel();
    stop(0);
}

void
doBaseBuildAmmo(GPlanet& p)
{
    WAmmoWindow aw(p);
    ui_root->centerWidget(aw);
    ui_root->add(aw);
    ui_root->run();
    ui_root->close(aw);
    getDisplayedTurn().doUpdates();
}
