/**
  *  \file client/dlg-mbox.cc
  */

#include "client/dlg-mbox.h"
#include "ui/util.h"
#include "cpluslib/format.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/commands.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "cpluslib/string.h"
#include "ui/msgbox.h"
#include "client/dlg-receiver.h"
#include "client/screens.h"

// FIXME: there should be a way to clear the when user leaves his race
static string_t current_search_str;

/**************************** WMessageDisplay ****************************/

/** \class WMessageDisplay
    \brief Widget displaying contents of a mailbox

    This widget displays one message of a mailbox. It automatically
    updates when the user chooses another message. */

WMessageDisplay::WMessageDisplay(int id, GMessagebox& mbox)
    : UIWidget(id), mbox(mbox)
{
    cookie = mbox.mbox_changed.add(static_cast<UIWidget*>(this),
                                   (void (UIWidget::*)()) &UIWidget::drawWidget);
}

WMessageDisplay::~WMessageDisplay()
{
    mbox.mbox_changed.remove(cookie);
}

void
WMessageDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_Flexible;
    info.min_size = info.pref_size = GfxPoint(41 * font_ems[FONT_FIXED],
                                              21 * font_heights[FONT_FIXED]);
}

void
WMessageDisplay::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UISkin& s = getSkin();
    s.startDraw();
    s.getParameters().font = fonts[FONT_FIXED];
    s.getParameters().bgcolor = GFX_TRANSPARENT;
    s.setColor(UIColor::tc_White);  // FIXME?
    s.drawBackground(can, getExtent());
    outTextFormattedF(can, s, getExtent().x, getExtent().y,
                      getExtent().w, getExtent().h / font_heights[FONT_FIXED],
                      mbox.getText(mbox.getCurrent()));

    string_t indicator = format(" %d/%d", mbox.getCurrent()+1, mbox.getCount());
    int indi_width = fonts[FONT_NORMAL]->getTextWidth(indicator);

    s.outText(can, FONT_NORMAL,
              getExtent().x + getExtent().w - indi_width,
              getExtent().y,
              indicator);

    ui.update(getExtent());
}

void
WMessageDisplay::next(int)
{
    mbox.setCurrent(mbox.getCurrent()+1);
}

void
WMessageDisplay::previous(int)
{
    mbox.setCurrent(mbox.getCurrent()-1);
}

void
WMessageDisplay::searchFromMessage(int nr)
{
    string_t substr = strUCase(current_search_str);
    while (nr < mbox.getCount()) {
        if (strUCase(mbox.getText(nr)).find(substr) != string_t::npos) {
            mbox.setCurrent(nr);
            return;
        }
        ++nr;
    }
    messageBox(_("Search text not found."), _("Search in messages"));
}

void
WMessageDisplay::search(int)
{
    if (UIInputLine::run(current_search_str, 40, _("Search in messages"),
                         _("Search for:"), 20, 0)
        && current_search_str.length())
    {
        searchFromMessage(0);
    }
}

void
WMessageDisplay::searchNext(int)
{
    if (current_search_str.length())
        searchFromMessage(mbox.getCurrent()+1);
    else
        search(0);
}

void
WMessageDisplay::forwardMessage()
{
    WReceiverSelector sel(0);
    if (sel.doSimpleDialog(_("Forward Message"))) {
        // FIXME: doSendMessage returns a screen to go to.
        doSendMessage("--- Forwarded Message ---\n" + mbox.getText(mbox.getCurrent()) + "\n--- End Forwarded Message ---",
                      sel.getReceivers());
    }        
}

void
WMessageDisplay::doStandardDialog()
{
    UIKeyTranslator::Init key_data[] = {
        SDLK_UP,       cm_Prev,
        SDLK_PAGEUP,   cm_Prev,
        SDLK_DOWN,     cm_Next,
        SDLK_PAGEDOWN, cm_Next
    };

    UIWindow win(_("Messages"), UIWindow::BLUE_WINDOW, COLOR_BLACK);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(*this);

    UIGroup g(UIHBoxLayout::instance5);
    win.add(g);

    UIButton prev(cm_Prev,  "\036", '-',          UIButton::aNothing);
    UIButton next(cm_Next,  "\037", '+',          UIButton::aNothing);
    UIButton search(cm_Next,  "S",    's',        UIButton::aNothing);
    UIButton snext (cm_Next,  "N",    'n',        UIButton::aNothing);
    UIButton forwd (cm_Next,  "F",    'f',        UIButton::aNothing);
    UIButton ok  (cm_Enter, "Close", SDLK_ESCAPE, UIButton::aStop);
    UIKeyTranslator keys(key_data, countof(key_data));
    UISpacer sp;

    g.add(prev);
    g.add(next);
    g.add(search);
    g.add(snext);
    g.add(forwd);
    g.add(sp);
    g.add(ok);
    g.add(keys);

    next.sig_fired.add(this, &WMessageDisplay::next);
    prev.sig_fired.add(this, &WMessageDisplay::previous);
    search.sig_fired.add(this, &WMessageDisplay::search);
    snext.sig_fired.add(this, &WMessageDisplay::searchNext);
    forwd.sig_fired.add(this, &WMessageDisplay::forwardMessage);

    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    ui_root->run();
    ui_root->close(win);
}
