/**
  *  \file client/dlg-supplies.cc
  *  \brief Supply Sale Dialog
  *
  *  This module contains the user interface for the "sell supplies"
  *  function.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. */

#include "client/actions.h"
#include "client/dlg-supplies.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "game/global.h"
#include "game/action.h"

/************************** WSellSuppliesDialog **************************/

/** \class WSellSuppliesDialog
    \brief Supply Sale Dialog

    This is the standard "sell supplies" dialog.

    When run, and the user chose "OK", the dialog will call
    GPlanetSellSuppliesAction::setAmount() and return nonzero; the
    caller must commit the action. When the user chooses "Cancel",
    this will return zero; the caller must cancel the transaction. */

/** Constructor. Constructs a complete dialog. */
WSellSuppliesDialog::WSellSuppliesDialog(GPlanetSellSuppliesAction& act)
    : UIWindow(_("Sell Supplies"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      select(100, 0, act.getMaximum(), 10), action(act)
{
    setLayout(UIVBoxLayout::instance5);
    add(holder.add(new UIMultilineStatic(FONT_NORMAL,
                                    format(_("You have %d kt supplies. You'll get 1 mc per kiloton. "
                                             "Remember that PCC automatically sells supplies when needed.\n"
                                             "Enter amount to sell:"),
                                           act.getMaximum()),
                                    400)));
    add(select);

    UIGroup& g = holder.add(new UIGroup(UIHBoxLayout::instance5));
    add(g);

    UIButton& ok = holder.add(new UIButton(1000, _("OK"), SDLK_RETURN, UIButton::aNothing));
    ok.sig_fired.add(this, &WSellSuppliesDialog::onEnter);
    g.add(ok);

    UIButton& all = holder.add(new UIButton(1001, UIKeyString(_("All but...")), UIButton::aNothing));
    all.sig_fired.add(this, &WSellSuppliesDialog::onAll);
    g.add(all);

    UIButton& esc = holder.add(new UIButton(1002, _("Cancel"), SDLK_ESCAPE, UIButton::aNothing));
    esc.sig_fired.add(this, &WSellSuppliesDialog::onCancel);
    g.add(esc);

    g.add(holder.add(new UISpacer()));

    pack();
}

WSellSuppliesDialog::~WSellSuppliesDialog()
{ }

/** Handler for "OK" button. */
void
WSellSuppliesDialog::onEnter(int)
{
    action.setAmount(select.getValue());
    stop(true);
}

/** Handler for "Sell all but" button. */
void
WSellSuppliesDialog::onAll(int)
{
    if (select.getValue() == 0 && action.getMaximum() != 0)
        if (!UIMessageBox(_("Do you really want to sell all supplies?"), _("Sell Supplies")).doYesNoDialog())
            return;
    action.setAmount(action.getMaximum() - select.getValue());
    stop(true);
}

/** Handler for "Cancel" button. */
void
WSellSuppliesDialog::onCancel(int)
{
    stop(false);
}

/** Main entry. Constructs the dialog, runs it, and applies it to
    planet p. */
void
WSellSuppliesDialog::doDialog(GPlanet& p)
{
    if (p.getCargo(el_Supplies) == 0) {
        messageBox(_("You do not have any supplies on this planet."),
                   _("Sell Supplies"));
        return;
    }
    
    GPlanetSellSuppliesAction act(p);
    WSellSuppliesDialog dlg(act);

    ui_root->centerWidget(dlg);
    ui_root->add(dlg);
    int i = dlg.run();
    ui_root->close(dlg);

    if (i)
        act.commit();
    else
        act.cancel();

    getDisplayedTurn().doUpdates();
}
