/**
  *  \file game/action.h
  *  \brief Game Actions
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_ACTION_H
#define PCC_V2_GAME_ACTION_H

#include "game/planet.h"
#include "game/cargo.h"
#include "game/struct.h"

class GObjectList;

/** \defgroup actions Game Actions

    These classes encapsulate game actions. They are used by the GUI
    code as well as the interpreter.

    Only nontrivial actions get their own class. Some actions have
    complicated preconditions, or interact with different objects;
    such things need an action class. Just changing a mission or
    friendly code can be done directly. */
//@{

class GPlanetBuildStarbaseAction : public GCargoCostTransaction {
    GPlanet&  the_planet;
    bool      want_build;
 public:
    GPlanetBuildStarbaseAction(GPlanet& the_planet, GCargoContainer& container,
                               bool want_build);
    ~GPlanetBuildStarbaseAction();

    void    commit();
};

class GPlanetBuildStructuresAction : public GCargoCostTransaction {
    enum { MAX = str_BaseDefense + 1 };
    struct Data {
        int   order;
        int   max;
        int   min;
    };
    Data data[MAX];
    GPlanet& the_planet;

    void computeCost();
 public:
    GPlanetBuildStructuresAction(GPlanet& the_planet, GCargoContainer& container);
    ~GPlanetBuildStructuresAction();

    int addStructures(TPlanetaryStructure ps, int count, bool partial);
    int addStructuresLimitCash(TPlanetaryStructure ps, int count);

    void commit();

    GPlanet& getPlanet() { return the_planet; }

    int getMin(TPlanetaryStructure ps) const { return data[ps].min; }
    int getMax(TPlanetaryStructure ps) const { return data[ps].max; }
    int getAmount(TPlanetaryStructure ps) const { return data[ps].order; }
};

class GPlanetSellSuppliesAction : public GTransaction {
    GPlanet&         the_planet;
    GPlanetTransfer  transfer;
    int32            amount;
    int32            reserved_sup;
 public:
    GPlanetSellSuppliesAction(GPlanet& the_planet);
    ~GPlanetSellSuppliesAction();

    int32 getMinimum() const;
    int32 getMaximum() const;

    void  setAmount(int32 amount);
    int32 getAmount() const { return amount; }

    void commit();
    void cancel();
};

class GStarbaseBuildAction : public GCargoCostTransaction {
    GPlanet&          the_planet;
 public:
    class Actor : public Object {
     public:
        virtual ~Actor();
        virtual void setTech(TTechLevel area, int value) = 0;
        virtual void setHullSlot(int index, int value) = 0;
        virtual void setEngine(int index, int value) = 0;
        virtual void setBeam(int index, int value) = 0;
        virtual void setLauncher(int index, int value) = 0;
        virtual void accountHull(int number, int count);
    };

    GStarbaseBuildAction(GPlanet& the_planet, GCargoContainer& container);
    ~GStarbaseBuildAction();

    void regenerate();
    
    void commit();
    void cancel();

    GPlanet& getPlanet() { return the_planet; }
 protected:
    virtual void doChanges(Actor& actor) = 0;
 private:
    class CountingActor;
    class ExecutingActor;
    friend class CountingActor;
    friend class ExecutingActor;
    void addItemCost(TCost c, int32 diff);
};

class GStarbaseTechUpgradeAction : public GStarbaseBuildAction {
    GPlanet& the_planet;
    int      new_tech[4], min_tech[4];

    void doChanges(Actor& actor);
 public:
    GStarbaseTechUpgradeAction(GPlanet& the_planet, GCargoContainer& container);
    ~GStarbaseTechUpgradeAction();

    bool     isValid(int index, int new_val) const;
    bool     setTech(int index, int new_val);
    bool     isValid() const { return GStarbaseBuildAction::isValid(); }
    int      getCost() const { return GStarbaseBuildAction::getCost(el_Money); }

    int      getTech(int index) const    { return new_tech[index]; }
    int      getMinTech(int index) const { return min_tech[index]; }
    int      getMaxTech(int index) const;

    int      getMissingCash() const { return getMissing(el_Money); }
    int      getReservedCash() const { return getReserved(el_Money); }
    int      getReservedSupplies() const { return getReserved(el_Supplies); }
};

class GStarbaseFixRecycleAction {
    int      num_own, num_foreign;
    GPlanet& the_planet;
 public:
    GStarbaseFixRecycleAction(GPlanet& the_planet);
    bool canFix() const;
    bool canRecycle() const;
    bool setAction(int action, int ship_id);
    void enumerateShipsFor(int action, GObjectList& list);
};

class GStarbaseBuildAmmoAction : public GCargoCostTransaction {
    enum { MAX = NUM_TORPS + 1 };
    GPlanetTransfer& pay;
    GCargoContainer& receiver;
    int32  added[MAX];
 public:
    GStarbaseBuildAmmoAction(GPlanetTransfer& pay, GCargoContainer& receiver);
    ~GStarbaseBuildAmmoAction();

    /* These do not use cargo types, they use 1..10 [torps] or 11. */
    bool   canHave(int index) const;
    int32  getAmount(int index) const;
    void   setAmount(int index, int32 amount);
    int    getTechUpgradeNeeded(int index) const;

    void   commit();
    void   cancel();

    GPlanet& getPlanet() { return pay.getPlanet(); }

    static GCargoType getCargoTypeFromIndex(int index);
};

class GStarbaseBuildShipAction : public GStarbaseBuildAction {
    TBuildOrder        order;
    bool               use_parts_from_storage : 1;
    bool               use_tech_upgrade       : 1;
    bool               is_ok                  : 1;

    void  doTechUpgrade(TTechLevel tl, Actor& a, int need);
    int   getBuildAmount(int need, int have);
 public:
    GStarbaseBuildShipAction(GPlanet& planet, GPlanetTransfer& pay);
    ~GStarbaseBuildShipAction();

    bool  isUsePartsFromStorage() const { return use_parts_from_storage; }
    bool  isUseTechUpgrade() const { return use_tech_upgrade; }

    void  setUsePartsFromStorage(bool b);
    void  setUseTechUpgrade(bool b);

    bool  isValid() const;

    void  commit();

    void                setBuildOrder(const TBuildOrder& o);
    const TBuildOrder&  getBuildOrder() const { return order; }

    void  doChanges(Actor& a);
};

//@}

#endif
