/**
  *  \file spec.h
  *
  *  Specification files. This file contains the classes which
  *  correspond to static game information (hullspec etc.)
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_SPEC_H
#define PCC_V2_SPEC_H

#include <string>
#include <vector>
#include "game/struct.h"
#include "io/io.h"
#include "cpluslib/assert.h"
#include "cpluslib/buffer.h"

/* Starship Component */
class GComponent {
    int id;
 public:
    GComponent(int aid) : id(aid) { }
    virtual ~GComponent();

    /** Get mass of this component, in kilotons. */
    virtual int getMass() const = 0;
    /** Get tech level required to build this component. */
    virtual int getTech() const = 0;
    /** Get cost of one of this components. */
    virtual TCost getCost() const = 0;
    /** Get name of this component. This is the full name. */
    virtual string_t  getName() const = 0;

    /** Get Id number of this component. */
    int getId() const { return id; }
};

/* Weapon */
class GWeapon : public GComponent {
 public:
    GWeapon(int aid) : GComponent(aid) { }
    /** Get kill (anti-life) power. Should not be 'AC-mangled'. */
    virtual int getKill() const = 0;
    /** Get damage (anti-structuer) power. */
    virtual int getDamage() const = 0;
};

/* Beam */
class GBeam : public GWeapon {
    TBeam data;
 public:
    GBeam(int aid, const TBeam& adata);

    string_t getName() const { return getBasicString(data.name); }
    int getKill() const { return data.kill; }
    int getDamage() const { return data.damage; }
    int getMass() const { return data.mass; }
    int getTech() const { return data.tech; }
    TCost getCost() const { return data.cost; }

    /** Get damage value. This is a fast access path bypassing vtbls. */
    int f_getDamage() const { return data.damage; }
    /** Get kill value. This is a fast access path bypassing vtbls. */
    int f_getKill() const { return data.kill; }
};

/* Torpedo Launcher */
class GTorpLauncher : public GWeapon {
    TTorpedo data;
 public:
    GTorpLauncher(int aid, const TTorpedo& adata);

    string_t getName() const { return getBasicString(data.name); }
    int getKill() const { return data.kill; }
    int getDamage() const { return data.damage; }
    int getMass() const { return data.laun_mass; }
    int getTech() const { return data.tech; }
    TCost getCost() const { return data.laun_cost; }

    /** Get cost of one torpedo in mc. */
    int getTorpCost() const { return data.torp_cost; }

    int f_getDamage() const { return data.damage; }
    int f_getKill() const { return data.kill; }
};

/* Engine */
class GEngine : public GComponent {
    TEngine data;
 public:
    GEngine(int aid, const TEngine& adata);

    int getMass() const { return 0; }
    int getTech() const { return data.tech; }
    TCost getCost() const { return data.cost; };
    string_t getName() const { return getBasicString(data.name); }

    int getFuelFactor(int warp) const;
};

enum {
    hf_MerlinAlchemy,             ///< 0 = Merlin Alchemy: 9 Sup -> 3 Min
    hf_NeutronicRefinery,         ///< 1 = Neutronic Refinery: Min + Sup -> Fuel
    hf_AriesRefinery,             ///< 2 = Aries Refinery: Min -> Fuel
    hf_HeatsTo50,                 ///< 3 = Bohemian Terraformer
    hf_CoolsTo50,                 ///< 4 = Eros Terraformer
    hf_HeatsTo100,                ///< 5 = Onyx Terraformer
    hf_Hyperdrive,                ///< 6
    hf_Gravitonic,                ///< 7
    hf_ScansAllWormholes,         ///< 8 = Bohemian
    hf_LadyRoyale,                ///< 9 = Lady Royale
    hf_LokiAnticloak,             ///< 10 = Loki
    hf_ImperialAssault,           ///< 11 = SSD
    hf_FirecloudChunnel,          ///< 12 = Firecloud
    hf_Ramscoop,                  ///< 13 = Cobol
    hf_FullBioscan,               ///< 14 = Pawn
    hf_AdvancedCloak,             ///< 15 = Dark Wing
    hf_Cloak,                     ///< 16
    hf_Bioscan,                   ///< 17
    hf_SaberGlory,                ///< 18 = Saber (10% damage to own ships)
    hf_D19bGlory,                 ///< 19 = D19b (20% damage to own ships)

    hf_MAX
};

class GHullFunction {
    string_t   name;
    string_t   description;
    int        id;
    int        implies;
 public:
    GHullFunction(int id, string_t name);
    ~GHullFunction();

    /** Get number of device which this one implies. -1 if none. */
    int getImpliedDevice() const { return implies; }

    /** Get device Id (hf_XXX). */
    int getId() const { return id; }

    /** Get device name. */
    const string_t& getName() const { return name; }

    string_t getDescription() const;

    void setDescription(string_t name);
    void setImpliedDevice(int implies);

    static void setDefaultSpecials();

    static GHullFunction* getHullFunctionById(int id);
    static GHullFunction* getHullFunctionByName(string_t name, bool partial = false);
    static GHullFunction* addHullFunction(int id, string_t name);

    static void readHullfuncDefinitions();
};

/* Starship Hull */
class GHull : public GComponent {
    THull data;
    Buffer<int16> specials;
 public:
    GHull(int aid, const THull& adata);
    ~GHull();

    int getMass() const { return data.mass; }
    int getTech() const { return data.tech; }
    TCost getCost() const;
    string_t getName() const { return getBasicString(data.name); }
    string_t getShortName() const { return getName(); /* FIXME!!! */ }

    /** Get picture number. \todo specify some properties of this value. */
    int getPicture() const { return data.picture; }
    /** Get size of fuel tank. */
    int getFuel() const { return data.fuel; }
    /** Get size of normal crew on this ship. */
    int getCrew() const { return data.crew; }
    /** Get number of engines on this ship. */
    int getNumEngines() const { return data.engines; }
    /** Get size of cargo room on this ship. */
    int getCargo() const { return data.cargo; }
    /** Get number of bays on this ship. */
    int getNumBays() const { return data.bays; }
    /** Get maximum number of torpedo launchers on this ship. */
    int getMaxLaunchers() const { return data.launchers; }
    /** Get maximum number of beam weapons on this ship. */
    int getMaxBeams() const { return data.beams; }

    void setSpecialFunction(int hf, int16 players, bool assign_to_hull);
    void clearSpecialFunctions();
    bool canDoSpecial(int hf, int player, bool hull) const;
    int  getPlayersThatCan(int hf, bool hull) const;
};

extern std::vector<GBeam> beam_spec;
extern std::vector<GTorpLauncher> torp_spec;
extern std::vector<GEngine> eng_spec;
extern std::vector<GHull*> hull_spec;
extern TTruehull truehull_spec;

void initSpecification();

/** Get beam. bid = beam number (1..NUM_BEAMS) */
inline GBeam&
getBeam(int bid) {
    ASSERT(bid > 0 && bid <= NUM_BEAMS);
    return beam_spec[bid-1];
}

/** Get torpedo launcher. tid = torpedo number (1..NUM_TORPS) */
inline GTorpLauncher&
getTorp(int tid) {
    ASSERT(tid > 0 && tid <= NUM_TORPS);
    return torp_spec[tid-1];
}

/** Get engine. eid = engine number (1..NUM_ENGINES) */
inline GEngine&
getEngine(int eid) {
    ASSERT(eid > 0 && eid <= NUM_ENGINES);
    return eng_spec[eid-1];
}

/** Get hull. hid = hull number (1..getNumHulls()) */
inline GHull&
getHull(int hid) {
    ASSERT(hid > 0 && unsigned(hid) <= hull_spec.size());
    return *hull_spec[hid-1];
}

/** Get truehull slot. pid = player number, slot = 1..20 */
inline int
getTruehull(int pid, int slot) {
    ASSERT(pid > 0 && pid <= NUM_PLAYERS);
    ASSERT(slot > 0 && slot <= NUM_HULLS_PER_PLAYER);
    return truehull_spec.hulls[pid-1][slot-1];
}

/** Get number of hulls. */
inline int
getNumHulls() {
    return hull_spec.size();
}

int getTruehullSlot(int pid, int hid);

#endif
